from copy import deepcopy

import django
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.forms import generic_inlineformset_factory
from django.core.cache import cache
from django.db import connection
from django.test import TestCase
from django.urls import reverse
from django.utils.timezone import datetime, now, timedelta
from freezegun import freeze_time
from swapper import get_model_name, load_model

from openwisp_controller.config.tests.test_admin import TestImportExportMixin
from openwisp_controller.config.tests.utils import CreateDeviceGroupMixin
from openwisp_controller.geo.tests.utils import TestGeoMixin
from openwisp_users.tests.utils import TestMultitenantAdminMixin

from ...check.settings import CHECK_CLASSES
from ..admin import CheckInline, CheckInlineFormSet
from . import DeviceMonitoringTestCase, TestWifiClientSessionMixin

Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')
DeviceData = load_model('device_monitoring', 'DeviceData')
WifiClient = load_model('device_monitoring', 'WifiClient')
WifiSession = load_model('device_monitoring', 'WifiSession')
User = get_user_model()
Check = load_model('check', 'Check')
# needed for config.geo
Device = load_model('config', 'Device')
DeviceLocation = load_model('geo', 'DeviceLocation')
Location = load_model('geo', 'Location')
# model_name changes if swapped
check_model_name = get_model_name('check', 'Check').lower().replace('.', '-')
metric_model_name = get_model_name('monitoring', 'Metric').lower().replace('.', '-')


class TestAdmin(
    TestWifiClientSessionMixin, TestImportExportMixin, DeviceMonitoringTestCase
):
    """Test the additions of openwisp-monitoring to DeviceAdmin"""

    resources_fields = TestImportExportMixin.resource_fields
    resources_fields.append('monitoring_status')
    app_label = 'config'
    _device_params = {
        'group': '',
        'management_ip': '',
        'model': '',
        'os': '',
        'system': '',
        'notes': '',
        'config-TOTAL_FORMS': '0',
        'config-INITIAL_FORMS': '0',
        'config-MIN_NUM_FORMS': '0',
        'config-MAX_NUM_FORMS': '1',
        # devicelocation
        'devicelocation-TOTAL_FORMS': '0',
        'devicelocation-INITIAL_FORMS': '0',
        'devicelocation-MIN_NUM_FORMS': '0',
        'devicelocation-MAX_NUM_FORMS': '1',
        # deviceconnection
        'deviceconnection_set-TOTAL_FORMS': '0',
        'deviceconnection_set-INITIAL_FORMS': '0',
        'deviceconnection_set-MIN_NUM_FORMS': '0',
        'deviceconnection_set-MAX_NUM_FORMS': '1000',
        # command
        'command_set-TOTAL_FORMS': '0',
        'command_set-INITIAL_FORMS': '0',
        'command_set-MIN_NUM_FORMS': '0',
        'command_set-MAX_NUM_FORMS': '1000',
        # check
        f'{check_model_name}-content_type-object_id-TOTAL_FORMS': '0',
        f'{check_model_name}-content_type-object_id-INITIAL_FORMS': '0',
        f'{check_model_name}-content_type-object_id-MIN_NUM_FORMS': '0',
        f'{check_model_name}-content_type-object_id-MAX_NUM_FORMS': '1000',
    }

    def setUp(self):
        self._login_admin()

    def tearDown(self):
        cache.clear()

    def _login_admin(self):
        u = User.objects.create_superuser('admin', 'admin', 'test@test.com')
        self.client.force_login(u)

    def test_device_admin(self):
        dd = self.create_test_data()
        check = Check.objects.create(
            name='Ping check',
            check_type=CHECK_CLASSES[0][0],
            content_object=dd,
            params={},
        )
        url = reverse('admin:config_device_change', args=[dd.pk])
        response = self.client.get(url)
        self.assertContains(response, '<h2>Status</h2>')
        self.assertContains(response, '<h2>Charts</h2>')
        self.assertContains(response, '<h2>Checks</h2>')
        self.assertContains(response, 'Storage')
        self.assertContains(response, 'CPU')
        self.assertContains(response, 'RAM status')
        self.assertContains(response, 'AlertSettings')
        self.assertContains(response, 'Is healthy')
        self.assertContains(response, 'http://testserver/api')
        self.assertContains(response, check.name)
        self.assertContains(response, check.params)

    def test_dashboard_map_on_index(self):
        url = reverse('admin:index')
        response = self.client.get(url)
        self.assertContains(response, "geoJsonUrl: \'http://testserver/api")
        self.assertContains(response, "locationDeviceUrl: \'http://testserver/api")

    def test_wifisession_dashboard_chart_query(self):
        url = reverse('admin:index')
        with self.assertNumQueries(10):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        device_monitoring_app_label = WifiSession._meta.app_label
        for query in connection.queries:
            if query['sql'] == (
                f'SELECT COUNT("{device_monitoring_app_label}_wifisession"."id") '
                f'AS "active__count" FROM "{device_monitoring_app_label}_wifisession" '
                f'WHERE "{device_monitoring_app_label}_wifisession"."stop_time" IS NULL'
            ):
                break
        else:
            self.fail('WiFiSession dashboard query not found')

    def test_status_data(self):
        d = self._create_device(organization=self._create_org())
        data = self._data()
        data.update(
            {
                "dhcp_leases": [
                    {
                        "expiry": 1586943200,
                        "mac": "f2:f1:3e:56:d2:77",
                        "ip": "192.168.66.196",
                        "client_name": "MyPhone1",
                        "client_id": "01:20:f4:78:19:3b:38",
                    },
                ],
                "neighbors": [
                    {
                        "mac": "44:D1:FA:4B:00:02",
                        "ip": "fe80::9683:c4ff:fe02:c2bf",
                        "interface": "eth2",
                        "state": "REACHABLE",
                    }
                ],
            }
        )
        data['interfaces'].append(deepcopy(data['interfaces'][0]))
        data['interfaces'][2].update(
            {
                'name': 'wlan2',
                'mac': '44:d1:fa:4b:38:45',
            }
        )
        data['interfaces'][2]['wireless']['mode'] = 'station'
        self._post_data(d.id, d.key, data)
        url = reverse('admin:config_device_change', args=[d.pk])
        r = self.client.get(url)
        with self.subTest('DHCP lease MAC is shown'):
            self.assertContains(r, 'f2:f1:3e:56:d2:77')
        with self.subTest('DHCP lease IP is shown'):
            self.assertContains(r, '192.168.66.196')
        with self.subTest('Neighbor MAC is shown'):
            self.assertContains(r, '44:D1:FA:4B:00:02')
        with self.subTest('Neighbor IP is shown'):
            self.assertContains(r, 'fe80::9683:c4ff:fe02:c2bf')
        with self.subTest('Wireless client table header is shown'):
            self.assertContains(
                r,
                '<th class="mac">\n\nAssociated client\n\nMAC address\n\n' '</th>',
                html=True,
                count=2,
            )
            self.assertContains(
                r,
                '<th class="mac">\n\nAccess Point\n\nMAC address\n\n' '</th>',
                html=True,
                count=1,
            )
        with self.subTest('Wireless interface properties are shown'):
            self.assertContains(
                r,
                '<div class="form-row">\n<label>Quality:</label>\n'
                '<div class="readonly">\n65 / 70\n</div>\n</div>',
                html=True,
            )
            self.assertContains(
                r,
                '<div class="form-row">\n<label>Bitrate:</label>\n'
                '<div class="readonly">\n1.1 MBits/s\n</div>\n</div>',
                html=True,
            )

    def test_status_data_contains_wifi_version(self):
        data = self._data()
        d = self._create_device(organization=self._create_org())
        url = reverse('admin:config_device_change', args=[d.pk])
        self._post_data(d.id, d.key, data)
        response = self.client.get(url)
        self.assertContains(
            response,
            """
                <div class="readonly">
                    WiFi 4 (802.11n): HT20
                </div>
            """,
            html=True,
        )
        self.assertContains(
            response,
            """
                <div class="readonly">
                    WiFi 5 (802.11ac): VHT80
                </div>
            """,
            html=True,
        )

    def test_status_data_contains_wifi_client_he_vht_ht_unknown(self):
        data = deepcopy(self._data())
        d = self._create_device(organization=self._create_org())
        url = reverse('admin:config_device_change', args=[d.pk])
        wireless_interface = data['interfaces'][0]['wireless']
        client = data['interfaces'][0]['wireless']['clients'][0]

        with self.subTest('Test when htmode is NOHT'):
            wireless_interface.update({'htmode': 'NOHT'})
            self._post_data(d.id, d.key, data)
            response = self.client.get(url)
            # make sure 'he', 'vht' and 'ht' are set to None
            self.assertContains(
                response,
                """
                    <td class="he">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                    <td class="vht">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                    <td class="ht">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                """,
                html=True,
            )

        with self.subTest('Test when htmode is HT and client he and vht are False'):
            wireless_interface.update({'htmode': 'HT40'})
            # both 'he' and 'vht' are False
            self.assertEqual(client['he'], False)
            self.assertEqual(client['vht'], False)
            self._post_data(d.id, d.key, data)
            response = self.client.get(url)
            # make sure 'he' and 'vht' are set to None
            self.assertContains(
                response,
                """
                    <td class="he">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                    <td class="vht">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                    <td class="ht">
                        <img src="/static/admin/img/icon-yes.svg">
                    </td>
                """,
                html=True,
            )

        with self.subTest('Test when htmode is VHT and client he is False'):
            wireless_interface.update({'htmode': 'VHT80'})
            client.update({'vht': True})
            self.assertEqual(client['he'], False)
            self.assertEqual(client['vht'], True)
            self._post_data(d.id, d.key, data)
            response = self.client.get(url)
            # make sure only 'he' is set to None
            self.assertContains(
                response,
                """
                    <td class="he">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                    <td class="vht">
                        <img src="/static/admin/img/icon-yes.svg">
                    </td>
                    <td class="ht">
                        <img src="/static/admin/img/icon-yes.svg">
                    </td>
                """,
                html=True,
            )

    def test_no_device_data(self):
        d = self._create_device(organization=self._create_org())
        url = reverse('admin:config_device_change', args=[d.pk])
        r = self.client.get(url)
        self.assertNotContains(r, '<h2>Status</h2>')
        self.assertNotContains(r, 'AlertSettings')

    def test_device_add_view(self):
        url = reverse('admin:config_device_add')
        r = self.client.get(url)
        self.assertNotContains(r, 'AlertSettings')
        self.assertContains(r, '<h2>Configuration</h2>')
        self.assertContains(r, '<h2>Map</h2>')
        self.assertContains(r, '<h2>Credentials</h2>')

    def test_device_disabled_organization_admin(self):
        self.create_test_data()
        device = Device.objects.first()
        Check.objects.create(
            name='Ping check',
            check_type=CHECK_CLASSES[0][0],
            content_object=device,
            params={},
        )
        org = device.organization
        org.is_active = False
        org.save()
        url = reverse('admin:config_device_change', args=[device.pk])
        response = self.client.get(url)
        self.assertContains(response, '<h2>Status</h2>')
        self.assertContains(response, '<h2>Charts</h2>')
        self.assertNotContains(response, '<h2>Checks</h2>')
        self.assertNotContains(response, '<h2>AlertSettings</h2>')

    def test_remove_invalid_interface(self):
        d = self._create_device(organization=self._create_org())
        dd = DeviceData(name='test-device', pk=d.pk)
        self._post_data(
            d.id,
            d.key,
            {'type': 'DeviceMonitoring', 'interfaces': [{'name': 'br-lan'}]},
        )
        url = reverse('admin:config_device_change', args=[dd.pk])
        self.client.get(url)

    def test_wifi_clients_admin(self):
        dd = self.create_test_data(no_resources=True)
        url = reverse('admin:config_device_change', args=[dd.id])
        r1 = self.client.get(url, follow=True)
        self.assertEqual(r1.status_code, 200)
        self.assertContains(r1, '00:ee:ad:34:f5:3b')

    def test_interface_properties_admin(self):
        dd = self.create_test_data(no_resources=True)
        url = reverse('admin:config_device_change', args=[dd.id])
        r1 = self.client.get(url, follow=True)
        self.assertEqual(r1.status_code, 200)
        self.assertContains(r1, '44:d1:fa:4b:38:44')
        self.assertContains(r1, 'Transmit Queue Length')
        self.assertContains(r1, 'Up')
        self.assertContains(r1, 'Multicast')
        self.assertContains(r1, 'MTU')

    def test_interface_bridge_admin(self):
        d = self._create_device(organization=self._create_org())
        dd = DeviceData(name='test-device', pk=d.pk)
        data = self._data()
        del data['resources']
        self._post_data(
            d.id,
            d.key,
            {
                'type': 'DeviceMonitoring',
                'interfaces': [
                    {
                        'name': 'br-lan',
                        'type': 'bridge',
                        'bridge_members': ['tap0', 'wlan0', 'wlan1'],
                        'stp': True,
                    }
                ],
            },
        )
        url = reverse('admin:config_device_change', args=[dd.id])
        r1 = self.client.get(url, follow=True)
        self.assertEqual(r1.status_code, 200)
        self.assertContains(r1, 'Bridge Members')
        self.assertContains(
            r1,
            """
                <a class="bridge-member"
                   href="javascript:scrollToElement('status-if-tap0')">
                    tap0
                </a>
            """,
            html=True,
        )
        self.assertContains(
            r1,
            """
                <a class="bridge-member"
                    href="javascript:scrollToElement('status-if-wlan0')">
                     wlan0
                </a>
            """,
            html=True,
        )
        self.assertContains(
            r1,
            """
                <a class="bridge-member"
                   href="javascript:scrollToElement('status-if-wlan1')">
                    wlan1
                </a>
            """,
            html=True,
        )
        self.assertContains(r1, 'Spanning Tree Protocol')

    def test_interface_mobile_admin(self):
        d = self._create_device(organization=self._create_org())
        self._post_data(
            d.id,
            d.key,
            {
                'type': 'DeviceMonitoring',
                'interfaces': [
                    {
                        'name': 'mobile0',
                        'mac': '00:00:00:00:00:00',
                        'mtu': 1900,
                        'multicast': True,
                        'txqueuelen': 1000,
                        'type': 'modem-manager',
                        'up': True,
                        'mobile': {
                            'connection_status': 'connected',
                            'imei': '300000001234567',
                            'manufacturer': 'Sierra Wireless, Incorporated',
                            'model': 'MC7430',
                            'operator_code': '50502',
                            'operator_name': 'YES OPTUS',
                            'power_status': 'on',
                            'signal': {
                                'lte': {'rsrp': -75, 'rsrq': -8, 'rssi': -51, 'snr': 13}
                            },
                        },
                    }
                ],
            },
        )
        url = reverse('admin:config_device_change', args=[d.id])
        r1 = self.client.get(url, follow=True)
        self.assertEqual(r1.status_code, 200)
        self.assertContains(r1, 'Signal Strength (LTE)')
        self.assertContains(r1, 'Signal Power (LTE)')
        self.assertContains(r1, 'Signal Quality (LTE)')
        self.assertContains(r1, 'Signal to noise ratio (LTE)')

    def test_uuid_bug(self):
        dd = self.create_test_data(no_resources=True)
        uuid = str(dd.pk).replace('-', '')
        url = reverse('admin:config_device_change', args=[uuid])
        r = self.client.get(url)
        self.assertContains(r, '<h2>Status</h2>')

    def test_check_inline_formset(self):
        d = self._create_device(organization=self._create_org())
        check_inline_formset = generic_inlineformset_factory(
            model=Check, form=CheckInline.form, formset=CheckInlineFormSet
        )
        ct = f'{check_model_name}-content_type-object_id'
        data = {
            f'{ct}-TOTAL_FORMS': '1',
            f'{ct}-INITIAL_FORMS': '0',
            f'{ct}-MAX_NUM_FORMS': '0',
            f'{ct}-0-name': 'Ping Check',
            f'{ct}-0-check_type': CHECK_CLASSES[0][0],
            f'{ct}-0-params': '{}',
            f'{ct}-0-is_active': True,
            f'{ct}-0-created': now(),
            f'{ct}-0-modified': now(),
        }
        formset = check_inline_formset(data)
        formset.instance = d
        self.assertTrue(formset.is_valid())
        self.assertEqual(formset.errors, [{}])
        self.assertEqual(formset.non_form_errors(), [])
        form = formset.forms[0]
        form.cleaned_data = data
        form.save(commit=True)
        self.assertEqual(Check.objects.count(), 1)
        c = Check.objects.first()
        self.assertEqual(c.name, 'Ping Check')
        self.assertEqual(c.content_object, d)

    def test_health_checks_list(self):
        dd = self.create_test_data()
        url = reverse('admin:config_device_change', args=[dd.pk])
        r = self.client.get(url)
        self.assertNotContains(r, '<label>Health checks:</label>')
        m = Metric.objects.filter(configuration='disk').first()
        m.write(m.alertsettings.threshold + 0.1)
        m.refresh_from_db()
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, False)
        self.assertEqual(dd.monitoring.status, 'problem')
        r = self.client.get(url)
        self.assertContains(r, '<label>Health checks:</label>')
        # Clients and Traffic metrics
        interface_metrics = dd.metrics.filter(alertsettings__isnull=True)
        interface_metric = interface_metrics.first()
        interface_metric.is_healthy = True
        interface_metric.save()
        for metric in interface_metrics:
            self.assertNotContains(r, f'{metric.name}</li>')
        other_metrics = dd.metrics.filter(alertsettings__isnull=False)
        for metric in other_metrics:
            health = 'yes' if metric.is_healthy else 'no'
            self.assertContains(
                r,
                f'<li><img src="/static/admin/img/icon-{health}.svg" '
                f'alt="health"> {metric.name}</li>',
            )

    def test_wifisession_inline(self):
        device = self._create_device()
        path = reverse('admin:config_device_change', args=[device.id])

        with self.subTest('Test inline absent when no WiFiSession is present'):
            response = self.client.get(path)
            self.assertNotContains(response, '<h2>WiFi Sessions</h2>')
            self.assertNotContains(response, 'monitoring-wifisession-changelist-url')

        wifi_session = self._create_wifi_session(device=device)

        with self.subTest('Test inline present when WiFiSession is open'):
            response = self.client.get(path)
            self.assertContains(response, '<h2>WiFi Sessions</h2>')
            self.assertContains(response, 'monitoring-wifisession-changelist-url')

        wifi_session.stop_time = now()
        wifi_session.save()

        with self.subTest('Test inline absent when WiFiSession is closed'):
            response = self.client.get(path)
            self.assertNotContains(response, '<h2>WiFi Sessions</h2>')
            self.assertNotContains(response, 'monitoring-wifisession-changelist-url')

    def test_check_alertsetting_inline(self):
        test_user = self._create_user(
            username='test', email='test@inline.com', is_staff=True
        )
        self._create_org_user(is_admin=True, user=test_user)
        device = self._create_device()
        ping_check = Check(
            check_type=CHECK_CLASSES[0][0], content_object=device, params={}
        )
        ping_check.full_clean()
        ping_check.save()
        url = reverse('admin:config_device_change', args=[device.pk])
        metric = self._create_general_metric(
            name='', content_object=device, configuration='ping'
        )
        self._create_alert_settings(metric=metric)
        self.client.force_login(test_user)

        def _add_device_permissions(user):
            user.user_permissions.clear()
            self.assertEqual(user.user_permissions.count(), 0)
            device_permissions = Permission.objects.filter(codename__endswith='device')
            # Permissions required to access device page
            user.user_permissions.add(*device_permissions),
            self.assertEqual(user.user_permissions.count(), 4)

        def _add_user_permissions(user, permission_query, expected_perm_count):
            user.user_permissions.add(*Permission.objects.filter(**permission_query))
            self.assertEqual(user.user_permissions.count(), expected_perm_count)

        def _assert_check_inline_in_response(response):
            self.assertContains(response, '<h2>Checks</h2>', html=True)
            self.assertContains(response, 'check-content_type-object_id-0-is_active')
            self.assertContains(response, 'check-content_type-object_id-0-check_type')
            self.assertContains(response, 'check-content_type-object_id-0-DELETE')

        def _assert_alertsettings_inline_in_response(response):
            self.assertContains(response, '<h2>Alert Settings</h2>', html=True)
            self.assertContains(response, 'form-row field-name')
            self.assertContains(
                response,
                '<img src="/static/admin/img/icon-yes.svg" alt="True">',
                html=True,
            )
            self.assertContains(response, '<h2>Advanced options</h2>', html=True)
            self.assertContains(
                response,
                'metric-content_type-object_id-0-alertsettings-0-is_active',
            )
            self.assertContains(
                response, '<option value="&lt;" selected>less than</option>'
            )
            self.assertContains(
                response,
                'metric-content_type-object_id-0-alertsettings-0-custom_threshold" value="1"',
            )
            self.assertContains(
                response,
                'metric-content_type-object_id-0-alertsettings-0-custom_tolerance" value="0"',
            )

        with self.subTest(
            'Test when a user does not have permission to access models or inline'
        ):
            _add_device_permissions(test_user)
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            self.assertNotContains(response, '<h2>Checks</h2>', html=True)
            self.assertNotContains(response, '<h2>Alert Settings</h2>', html=True)

        with self.subTest('Test check & alert settings with model permissions'):
            _add_device_permissions(test_user)
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            _add_user_permissions(test_user, {'codename__endswith': 'check'}, 8)
            _add_user_permissions(test_user, {'codename__endswith': 'metric'}, 12)
            _add_user_permissions(
                test_user, {'codename__endswith': 'alertsettings'}, 16
            )
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            _assert_check_inline_in_response(response)
            _assert_alertsettings_inline_in_response(response)

        with self.subTest('Test all inline permissions'):
            _add_device_permissions(test_user)
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            _add_user_permissions(test_user, {'codename__endswith': 'inline'}, 12)
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            _assert_check_inline_in_response(response)
            _assert_alertsettings_inline_in_response(response)

        with self.subTest('Test view inline permissions'):
            _add_device_permissions(test_user)
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            _add_user_permissions(
                test_user, {'codename__endswith': 'view_check_inline'}, 5
            )
            _add_user_permissions(
                test_user, {'codename__endswith': 'view_alertsettings_inline'}, 6
            )
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            self.assertContains(response, '<h2>Checks</h2>', html=True)
            self.assertContains(response, 'form-row field-check_type')
            self.assertContains(response, 'form-row field-is_active')
            self.assertContains(response, '<h2>Alert Settings</h2>', html=True)
            self.assertContains(response, 'form-row field-is_healthy djn-form-row-last')
            self.assertContains(
                response,
                '<img src="/static/admin/img/icon-yes.svg" alt="True">',
                html=True,
            )
            self.assertContains(
                response,
                'form-row field-is_active',
            )
            self.assertContains(response, 'form-row field-custom_operator')
            self.assertContains(
                response,
                'form-row field-custom_threshold',
            )
            self.assertContains(
                response,
                'form-row field-custom_tolerance',
            )

    def test_alert_settings_inline_post(self):
        device = self._create_device()
        metric = self._create_general_metric(
            name='', content_object=device, configuration='iperf3'
        )
        url = reverse('admin:config_device_change', args=[device.pk])
        alertsettings = self._create_alert_settings(metric=metric)
        test_inline_params = {
            'name': device.name,
            'organization': str(device.organization.id),
            'mac_address': device.mac_address,
            'key': device.key,
            # metric & alertsettings
            f'{metric_model_name}-content_type-object_id-TOTAL_FORMS': '1',
            f'{metric_model_name}-content_type-object_id-INITIAL_FORMS': '1',
            f'{metric_model_name}-content_type-object_id-MIN_NUM_FORMS': '0',
            f'{metric_model_name}-content_type-object_id-MAX_NUM_FORMS': '1000',
            f'{metric_model_name}-content_type-object_id-0-field_name': 'iperf3_result',
            f'{metric_model_name}-content_type-object_id-0-id': str(metric.id),
            f'{metric_model_name}-content_type-object_id-0-alertsettings-TOTAL_FORMS': '1',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-INITIAL_FORMS': '0',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-MIN_NUM_FORMS': '0',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-MAX_NUM_FORMS': '1',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-0-is_active': 'on',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_operator': '<',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_threshold': '9',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_tolerance': '1800',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-0-id': '',
            f'{metric_model_name}-content_type-object_id-0-alertsettings-0-metric': '',
        }
        # General metrics (clients & traffic) & Iperf3 are present
        self.assertEqual(Metric.objects.count(), 3)
        self.assertEqual(AlertSettings.objects.count(), 1)

        def _reset_alertsettings_inline():
            AlertSettings.objects.all().delete()

        # Delete AlertSettings objects before any subTests
        _reset_alertsettings_inline()
        # Delete all Metrics other than 'iperf3' before any subTests
        Metric.objects.exclude(configuration='iperf3').delete()

        def _assert_alertsettings_inline(response, operator, threshold, tolerance):
            self.assertEqual(response.status_code, 302)
            self.assertEqual(Metric.objects.count(), 1)
            self.assertEqual(AlertSettings.objects.count(), 1)
            alertsettings = AlertSettings.objects.first()
            self.assertEqual(alertsettings.operator, operator)
            self.assertEqual(alertsettings.threshold, threshold)
            self.assertEqual(alertsettings.tolerance, tolerance)

        with self.subTest('Test alert settings inline when all fields are provided'):
            self._device_params.update(test_inline_params)
            response = self.client.post(url, self._device_params)
            _assert_alertsettings_inline(response, '<', 9, 1800)
        _reset_alertsettings_inline()

        with self.subTest(
            'Test alert settings inline when partial fields are provided'
        ):
            test_inline_default_1 = {
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_operator': '>',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_threshold': '',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_tolerance': '',
            }
            self._device_params.update(test_inline_default_1)
            response = self.client.post(url, self._device_params)
            # 'threshold' and 'tolerance' are set to their default values
            _assert_alertsettings_inline(response, '>', 1, 0)
            _reset_alertsettings_inline()

            test_inline_default_2 = {
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_operator': '',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_threshold': '18',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_tolerance': '99',
            }
            self._device_params.update(test_inline_default_2)
            response = self.client.post(url, self._device_params)
            # 'operator' are set to their default values
            _assert_alertsettings_inline(response, '<', 18, 99)
        _reset_alertsettings_inline()

        with self.subTest('Test alert settings inline when all fields are absent'):
            test_inline_params_present = {
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_operator': '<',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_threshold': '99',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_tolerance': '1880',
            }
            self._device_params.update(test_inline_params_present)
            response = self.client.post(url, self._device_params)
            _assert_alertsettings_inline(response, '<', 99, 1880)

            alertsettings = AlertSettings.objects.first()
            metric = Metric.objects.first()

            test_inline_params_absent = {
                f'{metric_model_name}-content_type-object_id-INITIAL_FORMS': '1',
                f'{metric_model_name}-content_type-object_id-0-id': str(metric.id),
                f'{metric_model_name}-content_type-object_id-0-field_name': 'iperf3_result',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-INITIAL_FORMS': '1',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-id': str(
                    alertsettings.id
                ),
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-metric': str(
                    metric.id
                ),
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_operator': '',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_threshold': '',
                f'{metric_model_name}-content_type-object_id-0-alertsettings-0-custom_tolerance': '',
            }
            self._device_params.update(test_inline_params_absent)
            response = self.client.post(url, self._device_params)
            # If all the fields are empty, then it deletes the AlertSettings object
            # to prevent the default value from being used as a fallback
            self.assertEqual(response.status_code, 302)
            self.assertEqual(Metric.objects.count(), 1)
            self.assertEqual(AlertSettings.objects.count(), 0)

    def test_device_admin_recover_button_visibility(self):
        self._create_device(organization=self._create_org())
        url = reverse('admin:config_device_changelist')
        response = self.client.get(url)
        self.assertContains(
            response,
            """
                <li>
                    <a href="/admin/config/device/recover/" class="recoverlink">Recover deleted Devices</a>
                </li>
            """,
            html=True,
        )


class TestAdminDashboard(TestGeoMixin, DeviceMonitoringTestCase):
    location_model = Location
    object_location_model = DeviceLocation
    object_model = Device

    def test_dashboard(self):
        admin = User.objects.create_superuser('admin', 'admin', 'test@test.com')
        self.client.force_login(admin)
        self._create_object_location()
        response = self.client.get(reverse('admin:index'))
        static_files = [
            'monitoring/css/device-map.css',
            'monitoring/css/leaflet.fullscreen.css',
            'monitoring/css/netjsongraph.css',
            'leaflet/leaflet.css',
            'monitoring/js/lib/netjsongraph.min.js',
            'monitoring/js/lib/leaflet.fullscreen.min.js',
            'monitoring/js/device-map.js',
        ]
        for static_file in static_files:
            self.assertContains(response, static_file)
        self.assertContains(response, 'Monitoring Status')
        self.assertContains(response, '#267126')


class TestWifiSessionAdmin(
    CreateDeviceGroupMixin,
    TestMultitenantAdminMixin,
    TestWifiClientSessionMixin,
    TestCase,
):
    wifi_session_app_label = WifiSession._meta.app_label
    wifi_session_model_name = WifiSession._meta.model_name

    def setUp(self):
        admin = self._create_admin()
        self.client.force_login(admin)

    def test_changelist_filters_and_multitenancy(self):
        url = reverse(
            f'admin:{self.wifi_session_app_label}_{self.wifi_session_model_name}_changelist'
        )
        org1 = self._create_org(name='org1', slug='org1')
        org1_device_group = self._create_device_group(
            name='Org1 Routers', organization=org1
        )
        org1_device = self._create_device(
            name='org1-device', organization=org1, group=org1_device_group
        )
        org1_dd = self._create_device_data(device=org1_device)
        org1_interface_data = deepcopy(self._sample_data['interfaces'][0])
        org1_interface_data['name'] = 'org1_wlan0'
        org1_interface_data['wireless']['ssid'] = 'org1_wifi'
        org1_interface_data['wireless']['clients'][0]['mac'] = '00:ee:ad:34:f5:3b'

        org2 = self._create_org(name='org1', slug='org2')
        org2_device_group = self._create_device_group(
            name='Org2 Routers', organization=org2
        )
        org2_device = self._create_device(
            name='org2-device', organization=org2, group=org2_device_group
        )
        org2_dd = self._create_device_data(device=org2_device)
        org2_interface_data = deepcopy(self._sample_data['interfaces'][0])
        org2_interface_data['name'] = 'org2_wlan0'
        org2_interface_data['wireless']['ssid'] = 'org2_wifi'
        org2_interface_data['wireless']['clients'][0]['mac'] = '00:ee:ad:34:f5:3c'

        self._save_device_data(
            device_data=org1_dd,
            data={'type': 'DeviceMonitoring', 'interfaces': [org1_interface_data]},
        )
        self.assertEqual(WifiClient.objects.count(), 1)
        self.assertEqual(
            WifiSession.objects.filter(device__organization=org1).count(), 1
        )
        WifiSession.objects.filter(device__organization=org1).update(stop_time=now())

        with freeze_time(now() - timedelta(days=2)):
            self._save_device_data(
                device_data=org2_dd,
                data={'type': 'DeviceMonitoring', 'interfaces': [org2_interface_data]},
            )
        self.assertEqual(WifiClient.objects.count(), 2)
        self.assertEqual(
            WifiSession.objects.filter(device__organization=org2).count(), 1
        )

        def _assert_org2_wifi_session_in_response(
            response, org1_interface_data, org2_interface_data
        ):
            self.assertContains(
                response, '<p class="paginator">\n\n1 WiFi Session\n\n\n</p>'
            )
            self.assertContains(
                response,
                '<td class="field-ssid">{}</td>'.format(
                    org2_interface_data['wireless']['ssid']
                ),
            )
            self.assertContains(
                response, org2_interface_data['wireless']['clients'][0]['mac']
            )
            self.assertNotContains(
                response,
                '<td class="field-ssid">{}</td>'.format(
                    org1_interface_data['wireless']['ssid']
                ),
            )
            self.assertNotContains(
                response, org1_interface_data['wireless']['clients'][0]['mac']
            )

        with self.subTest('Test without filters'):
            response = self.client.get(url)
            self.assertContains(
                response, '<p class="paginator">\n\n2 WiFi Sessions\n\n\n</p>'
            )

        with self.subTest('Test start_time filter'):
            response = self.client.get(
                url, {'start_time__lte': now() - timedelta(days=1)}
            )
            _assert_org2_wifi_session_in_response(
                response, org1_interface_data, org2_interface_data
            )

        with self.subTest('Test stop_time filter'):
            response = self.client.get(url, {'stop_time__isnull': 'True'})
            _assert_org2_wifi_session_in_response(
                response, org1_interface_data, org2_interface_data
            )

        with self.subTest('Test organization filter'):
            response = self.client.get(
                url, {'device__organization__id__exact': str(org2.pk)}
            )
            _assert_org2_wifi_session_in_response(
                response, org1_interface_data, org2_interface_data
            )

        with self.subTest('Test device_group filter'):
            response = self.client.get(
                url, {'device__group__id__exact': str(org2_device_group.pk)}
            )
            _assert_org2_wifi_session_in_response(
                response, org1_interface_data, org2_interface_data
            )

        with self.subTest('Test device filter'):
            # Filter by device pk
            response = self.client.get(url, {'device': str(org2_device.pk)})
            _assert_org2_wifi_session_in_response(
                response, org1_interface_data, org2_interface_data
            )

        with self.subTest('Test multitenancy'):
            administrator = self._create_administrator(organizations=[org2])
            self.client.force_login(administrator)
            response = self.client.get(url)
            _assert_org2_wifi_session_in_response(
                response, org1_interface_data, org2_interface_data
            )

    def test_wifi_session_chart_on_index(self):
        url = reverse('admin:index')
        self._create_wifi_session()
        response = self.client.get(url)
        self.assertContains(response, 'Currently Active WiFi Sessions')
        self.assertContains(response, 'Open WiFi session list')

    def test_deleting_device_with_wifisessions(self):
        device_data = self._save_device_data()
        device = Device.objects.first()
        device.deactivate()
        path = reverse('admin:config_device_delete', args=[device_data.pk])
        response = self.client.post(path, {'post': 'yes'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Device.objects.count(), 0)

    def test_dashboard_wifi_session_chart(self):
        org1 = self._create_org(name='org1', slug='org1')
        org1_device = self._create_device(organization=org1)
        self._create_wifi_session(device=org1_device)
        org2 = self._create_org(name='org2', slug='org2')
        administrator = self._create_administrator([org2])
        self.client.force_login(administrator)
        response = self.client.get(reverse('admin:index'))
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.context['dashboard_charts'][13]['query_params'],
            {'labels': [], 'values': []},
        )

    def test_wifi_client_he_vht_ht_unknown(self):
        test_wifi_client = self._create_wifi_client(he=None, vht=None, ht=None)
        test_wifi_session = self._create_wifi_session(wifi_client=test_wifi_client)
        device = Device.objects.all().first()

        with self.subTest('Test device wifi session inline'):
            url = reverse('admin:config_device_change', args=[device.id])
            response = self.client.get(url)
            self.assertContains(
                response,
                """
                    <td class="field-he">
                        <p><img src="/static/admin/img/icon-unknown.svg"></p>
                    </td>
                    <td class="field-vht">
                        <p><img src="/static/admin/img/icon-unknown.svg"></p>
                    </td>
                    <td class="field-ht">
                        <p><img src="/static/admin/img/icon-unknown.svg"></p>
                    </td>
                """,
                html=True,
            )
        with self.subTest('Test device wifi session list'):
            url = reverse(
                f'admin:{self.wifi_session_app_label}_{self.wifi_session_model_name}_changelist'
            )
            response = self.client.get(url)
            self.assertContains(
                response,
                """
                    <td class="field-he">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                    <td class="field-vht">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                    <td class="field-ht">
                        <img src="/static/admin/img/icon-unknown.svg">
                    </td>
                """,
                html=True,
            )
        with self.subTest('Test device wifi session change'):
            url = reverse(
                f'admin:{self.wifi_session_app_label}_{self.wifi_session_model_name}_change',
                args=[test_wifi_session.id],
            )
            response = self.client.get(url)
            self.assertContains(
                response,
                """
                    <div class="form-row field-he">
                        <div>
                            {start_div}
                                <label>WiFi 6 (802.11ax):</label>
                                    <div class="readonly">
                                        <img src="/static/admin/img/icon-unknown.svg">
                                    </div>
                            {end_div}
                        </div>
                    </div>
                    <div class="form-row field-vht">
                        <div>
                            {start_div}<label>WiFi 5 (802.11ac):</label>
                                <div class="readonly">
                                    <img src="/static/admin/img/icon-unknown.svg">
                                </div>
                            {end_div}
                        </div>
                    </div>
                    <div class="form-row field-ht">
                        <div>
                            {start_div}<label>WiFi 4 (802.11n):</label>
                                <div class="readonly">
                                    <img src="/static/admin/img/icon-unknown.svg">
                                </div>
                            {end_div}
                        </div>
                    </div>
                """.format(
                    # TODO: Remove this when dropping support for Django 3.2 and 4.0
                    start_div=(
                        '<div class="flex-container">'
                        if django.VERSION >= (4, 2)
                        else ''
                    ),
                    end_div='</div>' if django.VERSION >= (4, 2) else '',
                ),
                html=True,
            )

    def test_wifi_session_stop_time_formatting(self):
        start_time = datetime.strptime('2023-8-24 17:08:00', '%Y-%m-%d %H:%M:%S')
        stop_time = datetime.strptime('2023-8-24 19:46:00', '%Y-%m-%d %H:%M:%S')
        session = self._create_wifi_session(stop_time=stop_time)
        # WifiSession.start_time has "auto_now" set to True.
        # To overwrite the current date, we set the start_time explicitly
        WifiSession.objects.update(start_time=start_time)
        url = reverse(
            f'admin:{self.wifi_session_app_label}_{self.wifi_session_model_name}_change',
            args=(session.id,),
        )
        response = self.client.get(url)
        self.assertContains(
            response,
            (
                '<label>Start time:</label>\n\n'
                '<div class="readonly">Aug. 24, 2023, 5:08 p.m.</div>'
            ),
            html=True,
        )
        self.assertContains(
            response,
            (
                '<label>Stop time:</label>\n\n'
                '<div class="readonly">Aug. 24, 2023, 7:46 p.m.</div>'
            ),
            html=True,
        )



# Generated by Django 4.0.4 on 2022-08-19 11:28

from django.db import migrations

from . import assign_check_inline_permissions_to_groups


class Migration(migrations.Migration):
    dependencies = [
        ('check', '0008_alter_check_options'),
    ]

    operations = [
        migrations.RunPython(
            assign_check_inline_permissions_to_groups,
            reverse_code=migrations.RunPython.noop,
        ),
    ]



import json
from datetime import date, timedelta
from unittest.mock import patch

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.test import TestCase
from django.utils.timezone import now
from swapper import load_model

from openwisp_utils.tests import capture_stderr

from .. import settings as app_settings
from ..configuration import (
    CHART_CONFIGURATION_CHOICES,
    DEFAULT_DASHBOARD_TRAFFIC_CHART,
    get_chart_configuration,
    register_chart,
    unregister_chart,
)
from . import TestMonitoringMixin, charts

Chart = load_model('monitoring', 'Chart')


class TestCharts(TestMonitoringMixin, TestCase):
    """Tests for functionalities related to charts"""

    def test_read(self):
        c = self._create_chart()
        data = self._read_chart(c)
        key = c.metric.field_name
        self.assertIn('x', data)
        self.assertIn('traces', data)
        self.assertEqual(len(data['x']), 3)
        charts = data['traces']
        self.assertEqual(charts[0][0], key)
        self.assertEqual(len(charts[0][1]), 3)
        self.assertEqual(charts[0][1], [3, 6, 9])

    def test_read_summary_avg(self):
        m = self._create_object_metric(name='summary_avg')
        c = self._create_chart(metric=m, test_data=False, configuration='mean_test')
        m.write(1, time=now() - timedelta(days=2))
        m.write(2, time=now() - timedelta(days=1))
        m.write(3, time=now())
        data = self._read_chart(c)
        self.assertEqual(data['summary'], {'value': 2})

    def test_read_summary_sum(self):
        m = self._create_object_metric(name='summary_sum')
        c = self._create_chart(metric=m, test_data=False, configuration='sum_test')
        m.write(5, time=now() - timedelta(days=2))
        m.write(4, time=now() - timedelta(days=1))
        m.write(1, time=now())
        data = self._read_chart(c)
        self.assertEqual(data['summary'], {'value': 10})

    def test_read_summary_not_aggregate(self):
        m = self._create_object_metric(name='summary_hidden')
        c = self._create_chart(metric=m)
        data = self._read_chart(c)
        self.assertEqual(data['summary'], {'value': None})

    def test_read_summary_top_fields(self):
        m = self._create_object_metric(
            name='applications', configuration='top_fields_mean'
        )
        c = self._create_chart(
            metric=m, test_data=False, configuration='top_fields_mean'
        )
        m.write(
            0,
            extra_values={
                'google': 150.00000001,
                'facebook': 0.00911111,
                'reddit': 0.0,
            },
            time=now() - timedelta(days=2),
        )
        m.write(
            0,
            extra_values={
                'google': 200.00000001,
                'facebook': 0.00611111,
                'reddit': 0.0,
            },
            time=now() - timedelta(days=1),
        )
        m.write(
            0,
            extra_values={
                'google': 250.00000001,
                'facebook': 0.00311111,
                'reddit': 0.0,
            },
            time=now() - timedelta(days=0),
        )
        data = self._read_chart(c)
        self.assertEqual(data['summary'], {'google': 200.0, 'facebook': 0.0061})

    def test_read_summary_top_fields_acid(self):
        m = self._create_object_metric(
            name='applications', configuration='top_fields_mean'
        )
        c = self._create_chart(
            metric=m, test_data=False, configuration='top_fields_mean'
        )
        m.write(
            0,
            extra_values={'google': 0.0, 'facebook': 6000.0, 'reddit': 0.0},
            time=now() - timedelta(days=3),
        )
        m.write(
            0,
            extra_values={
                'google': 150000000.0,
                'facebook': 90000000.0,
                'reddit': 1000.0,
            },
            time=now() - timedelta(days=2),
        )
        m.write(
            0,
            extra_values={'google': 200000000.0, 'facebook': 60000000.0, 'reddit': 0.0},
            time=now() - timedelta(days=1),
        )
        m.write(
            0,
            extra_values={'google': 0.0, 'facebook': 6000.0, 'reddit': 0.0},
            time=now() - timedelta(days=0),
        )
        data = self._read_chart(c)
        self.assertEqual(data['summary'], {'google': 87500000, 'facebook': 37503000})
        self.assertEqual(c.get_top_fields(2), ['google', 'facebook'])

    def test_read_multiple(self):
        c = self._create_chart(test_data=None, configuration='multiple_test')
        m1 = c.metric
        m2 = self._create_object_metric(
            name='test metric 2',
            key='test_metric',
            field_name='value2',
            content_object=m1.content_object,
        )
        now_ = now()
        for n in range(0, 3):
            time = now_ - timedelta(days=n)
            m1.write(n + 1, time=time)
            m2.write(n + 2, time=time)
        data = self._read_chart(c)
        f1 = m1.field_name
        f2 = 'value2'
        self.assertIn('x', data)
        self.assertIn('traces', data)
        self.assertEqual(len(data['x']), 3)
        charts = data['traces']
        self.assertIn(f1, charts[0][0])
        self.assertIn(f2, charts[1][0])
        self.assertEqual(len(charts[0][1]), 3)
        self.assertEqual(len(charts[1][1]), 3)
        self.assertEqual(charts[0][1], [3, 2, 1])
        self.assertEqual(charts[1][1], [4, 3, 2])

    def test_read_group_by_tag(self):
        m1 = self._create_object_metric(
            name='test metric 1',
            key='test_metric',
            field_name='item',
            extra_tags={'metric_num': 'first'},
        )
        m2 = self._create_object_metric(
            name='test metric 2',
            key='test_metric',
            field_name='item',
            content_object=self._create_org(),
            extra_tags={'metric_num': 'second'},
        )
        c = self._create_chart(metric=m1, test_data=False, configuration='group_by_tag')
        now_ = now()
        for n in range(0, 3):
            time = now_ - timedelta(days=n)
            m1.write(n + 1, time=time)
            m2.write(n + 2, time=time)
        data = self._read_chart(c, time='30d', start_date=time)
        self.assertIn('x', data)
        self.assertIn('traces', data)
        self.assertEqual(len(data['x']), 3)
        charts = data['traces']
        self.assertIn('first', charts[0][0])
        self.assertIn('second', charts[1][0])
        self.assertEqual(len(charts[0][1]), 3)
        self.assertEqual(len(charts[1][1]), 3)
        self.assertEqual(charts[0][1], [3, 5, 6])
        self.assertEqual(charts[1][1], [4, 7, 9])
        self.assertEqual(data['summary']['first'], 6)
        self.assertEqual(data['summary']['second'], 9)

    def test_json(self):
        c = self._create_chart()
        data = self._read_chart(c)
        # convert tuples to lists otherwise comparison will fail
        for i, chart in enumerate(data['traces']):
            data['traces'][i] = list(chart)
        # update data with unit
        data.update(
            {
                'unit': c.unit,
                'trace_type': c.trace_type,
                'trace_order': c.trace_order,
                'calculate_total': False,
                'connect_points': c.connect_points,
                'colors': c.colors,
            }
        )
        self.assertDictEqual(json.loads(c.json()), data)

    def test_read_bad_query(self):
        try:
            self._create_chart(configuration='bad_test')
        except ValidationError as e:
            self.assertIn('configuration', e.message_dict)
            self.assertIn('error parsing query: found BAD', str(e.message_dict))
        else:
            self.fail('ValidationError not raised')

    def test_get_query(self):
        c = self._create_chart(test_data=False)
        m = c.metric
        now_ = now()
        today = date(now_.year, now_.month, now_.day)
        time = today - timedelta(days=6)
        expected = c.query.format(
            field_name=m.field_name,
            key=m.key,
            content_type=m.content_type_key,
            object_id=m.object_id,
            time=str(time),
            end_date='',
        )
        expected = "{0} tz('{1}')".format(expected, settings.TIME_ZONE)
        self.assertEqual(c.get_query(), expected)

    def test_description(self):
        c = self._create_chart(test_data=False)
        self.assertEqual(c.description, 'Dummy chart for testing purposes.')

    def test_wifi_hostapd(self):
        m = self._create_object_metric(
            name='wifi associations',
            key='hostapd',
            field_name='mac',
            extra_tags={'ifname': 'wlan0'},
        )
        c = self._create_chart(metric=m, test_data=False, configuration='wifi_clients')
        now_ = now()
        for n in range(0, 9):
            m.write('00:16:3e:00:00:00', time=now_ - timedelta(days=n))
            m.write('00:23:4b:00:00:00', time=now_ - timedelta(days=n, seconds=1))
        m.write('00:16:3e:00:00:00', time=now_ - timedelta(days=2))
        m.write('00:16:3e:00:00:00', time=now_ - timedelta(days=4))
        m.write('00:23:4a:00:00:00')
        m.write('00:14:5c:00:00:00')
        c.save()
        data = self._read_chart(c, time='30d')
        self.assertEqual(data['traces'][0][0], 'wifi_clients')
        # last 10 days
        self.assertEqual(data['traces'][0][1][-10:], [0, 2, 2, 2, 2, 2, 2, 2, 2, 4])

    def test_get_time(self):
        c = Chart()
        now_ = now()
        today = date(now_.year, now_.month, now_.day)
        self.assertIn(str(today - timedelta(days=30)), c._get_time('30d'))
        self.assertIn(str(now() - timedelta(days=1))[0:10], c._get_time('1d'))
        self.assertIn(str(now() - timedelta(days=3))[0:10], c._get_time('3d'))

    def test_get_top_fields(self):
        m = self._create_object_metric(name='test', configuration='get_top_fields')
        c = self._create_chart(metric=m, test_data=None, configuration='histogram')
        self.assertEqual(c.get_top_fields(number=3), [])
        self._write_metric(
            m, None, extra_values={'http2': 100, 'ssh': 90, 'udp': 80, 'spdy': 70}
        )
        self.assertEqual(c.get_top_fields(number=3), ['http2', 'ssh', 'udp'])

    def test_query_histogram(self):
        m = self._create_object_metric(name='histogram', configuration='get_top_fields')
        self._write_metric(
            m, None, extra_values={'http2': 100, 'ssh': 90, 'udp': 80, 'spdy': 70}
        )
        c = Chart(metric=m, configuration='histogram')
        c.full_clean()
        c.save()
        self.assertNotIn('GROUP BY time', c.get_query())

    @capture_stderr()
    def test_bad_json_query_returns_none(self):
        m = self._create_object_metric(
            name='wifi associations',
            key='hostapd',
            field_name='mac',
            extra_tags={'ifname': 'wlan0'},
        )
        c = self._create_chart(metric=m, test_data=False, configuration='wifi_clients')
        m.write('00:14:5c:00:00:00')
        c.save()
        self.assertIsNone(c.json(time=1))

    def test_register_invalid_chart_configuration(self):
        with self.subTest('Registering with incomplete chart configuration.'):
            with self.assertRaises(AssertionError):
                register_chart('test_type', dict())
        with self.subTest('Registering already registered chart configuration.'):
            histogram = charts['histogram']
            register_chart('histogram_test', histogram)
            with self.assertRaises(ImproperlyConfigured):
                register_chart('histogram_test', histogram)
            unregister_chart('histogram_test')
        with self.subTest('Registering with improper chart configuration name'):
            with self.assertRaises(ImproperlyConfigured):
                register_chart(['test_type'], dict())
        with self.subTest('Registering with improper chart configuration'):
            with self.assertRaises(ImproperlyConfigured):
                register_chart('test_type', tuple())

    def test_unregister_invalid_chart_configuration(self):
        with self.subTest('Unregistering with improper chart configuration name'):
            with self.assertRaises(ImproperlyConfigured):
                unregister_chart(dict())
        with self.subTest('Unregistering unregistered chart configuration'):
            with self.assertRaises(ImproperlyConfigured):
                unregister_chart('test_chart')

    def test_register_valid_chart_configuration(self):
        dummy = charts['dummy']
        register_chart('dummy_test', dummy)
        self.assertIn(('dummy_test', 'dummy_test'), CHART_CONFIGURATION_CHOICES)
        unregister_chart('dummy_test')
        self.assertNotIn(('dummy_test', 'dummy_test'), CHART_CONFIGURATION_CHOICES)

    def test_additional_charts_setting(self):
        self.assertNotIn('dummy_test', get_chart_configuration())
        chart = {'dummy_test': charts['dummy']}
        with patch.object(app_settings, 'ADDITIONAL_CHARTS', chart):
            self.assertEqual(
                get_chart_configuration()['dummy_test'], chart['dummy_test']
            )

    def test_clean_dashboard_traffic_chart(self):
        with self.subTest('Test default dashboard traffic chart interfaces'):
            self.assertDictEqual(
                DEFAULT_DASHBOARD_TRAFFIC_CHART, {'__all__': ['wan', 'eth1', 'eth0_2']}
            )



# Generated by Django 3.1.2 on 2020-11-18 12:55

import collections
import uuid

import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import model_utils.fields
import swapper
from django.db import migrations, models

from ..settings import CHECK_CLASSES


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        swapper.dependency('config', 'Device', '0004_add_device_model'),
    ]

    operations = [
        migrations.CreateModel(
            name='Check',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                ('name', models.CharField(db_index=True, max_length=64)),
                ('active', models.BooleanField(db_index=True, default=True)),
                ('description', models.TextField(blank=True, help_text='Notes')),
                (
                    'object_id',
                    models.CharField(blank=True, db_index=True, max_length=36),
                ),
                (
                    'check',
                    models.CharField(
                        choices=CHECK_CLASSES,
                        db_index=True,
                        help_text='Select check type',
                        max_length=128,
                        verbose_name='check type',
                    ),
                ),
                (
                    'params',
                    jsonfield.fields.JSONField(
                        blank=True,
                        default=dict,
                        dump_kwargs={'indent': 4},
                        help_text='parameters needed to perform the check',
                        load_kwargs={'object_pairs_hook': collections.OrderedDict},
                        verbose_name='parameters',
                    ),
                ),
                (
                    'content_type',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to='contenttypes.contenttype',
                    ),
                ),
            ],
            options={
                'abstract': False,
                'swappable': swapper.swappable_setting('check', 'Check'),
                'unique_together': {('name', 'object_id', 'content_type')},
            },
        )
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse

from . import TestMonitoringMixin


class TestAdmin(TestMonitoringMixin, TestCase):
    def _login_admin(self):
        User = get_user_model()
        u = User.objects.create_superuser('admin', 'admin', 'test@test.com')
        self.client.force_login(u)

    def test_metric_admin(self):
        m = self._create_general_metric()
        url = reverse('admin:monitoring_metric_change', args=[m.pk])
        self._login_admin()
        r = self.client.get(url)
        self.assertEqual(r.status_code, 200)

    def test_alert_settings_inline(self):
        m = self._create_general_metric(configuration='ping')
        alert_s = self._create_alert_settings(metric=m)
        self.assertIsNone(alert_s.custom_operator)
        self.assertIsNone(alert_s.custom_threshold)
        self.assertIsNone(alert_s.custom_tolerance)
        url = reverse('admin:monitoring_metric_change', args=[m.pk])
        self._login_admin()
        r = self.client.get(url)
        self.assertEqual(r.status_code, 200)
        self.assertContains(r, '<option value="&lt;" selected>less than</option>')
        self.assertContains(r, 'name="alertsettings-0-custom_threshold" value="1"')
        self.assertContains(r, 'name="alertsettings-0-custom_tolerance" value="0"')

    def test_admin_menu_groups(self):
        # Test menu group (openwisp-utils menu group) for Metric and Check models
        self._login_admin()
        response = self.client.get(reverse('admin:index'))
        with self.subTest('test menu group link for check model'):
            url = reverse('admin:check_check_changelist')
            self.assertContains(response, f'class="mg-link" href="{url}"')
        with self.subTest('test menu group link for metric model'):
            url = reverse('admin:monitoring_metric_changelist')
            self.assertContains(response, f'class="mg-link" href="{url}"')
        with self.subTest('test "monitoring" group is registered'):
            self.assertContains(
                response,
                '<div class="mg-dropdown-label">Monitoring </div>',
                html=True,
            )



import logging
import uuid
from datetime import datetime

from cache_memoize import cache_memoize
from django.contrib.contenttypes.models import ContentType
from django.contrib.gis.db.models.functions import Distance
from django.core.exceptions import ValidationError
from django.db.models import Count, Q
from django.db.models.functions import Round
from django.http import Http404
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from pytz import UTC
from rest_framework import pagination, serializers, status
from rest_framework.generics import (
    GenericAPIView,
    ListAPIView,
    RetrieveAPIView,
    get_object_or_404,
)
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response
from swapper import load_model

from openwisp_controller.config.api.views import DeviceListCreateView
from openwisp_controller.geo.api.views import (
    DevicePermission,
    GeoJsonLocationList,
    LocationDeviceList,
    ProtectedAPIMixin,
)
from openwisp_users.api.mixins import FilterByOrganizationManaged

from ...settings import CACHE_TIMEOUT
from ...views import MonitoringApiViewMixin
from ..schema import schema
from ..signals import device_metrics_received
from ..tasks import write_device_metrics
from .filters import (
    MonitoringDeviceFilter,
    MonitoringNearbyDeviceFilter,
    WifiSessionFilter,
)
from .serializers import (
    MonitoringDeviceDetailSerializer,
    MonitoringDeviceListSerializer,
    MonitoringGeoJsonLocationSerializer,
    MonitoringLocationDeviceSerializer,
    MonitoringNearbyDeviceSerializer,
    WifiSessionSerializer,
)

logger = logging.getLogger(__name__)
Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')
Device = load_model('config', 'Device')
DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
DeviceData = load_model('device_monitoring', 'DeviceData')
Location = load_model('geo', 'Location')
WifiSession = load_model('device_monitoring', 'WifiSession')


class ListViewPagination(pagination.PageNumberPagination):
    page_size = 10
    page_size_query_param = 'page_size'
    max_page_size = 100


def get_device_args_rewrite(view, pk):
    """Use only the PK parameter for calculating the cache key"""
    try:
        pk = uuid.UUID(pk)
    except ValueError:
        return pk
    return pk.hex


def get_charts_args_rewrite(view, request, pk):
    return (pk,)


class DeviceKeyAuthenticationMixin(object):
    def get_permissions(self):
        if self.request.method in SAFE_METHODS and not self.request.query_params.get(
            'key'
        ):
            self.permission_classes = ProtectedAPIMixin.permission_classes
        return super().get_permissions()

    def get_authenticators(self):
        if self.request.method in SAFE_METHODS and not self.request.GET.get('key'):
            self.authentication_classes = ProtectedAPIMixin.authentication_classes
        return super().get_authenticators()


class DeviceMetricView(
    DeviceKeyAuthenticationMixin, MonitoringApiViewMixin, GenericAPIView
):
    """Device Monitoring View.

    Retrieve device information, monitoring status (health status), a list
    of metrics, chart data and optionally device status information (if
    ``?status=true``).

    Suports session authentication, token authentication, or alternatively
    device key authentication passed as query string parameter (this
    method is meant to be used by network devices).
    """

    model = DeviceData
    queryset = (
        DeviceData.objects.filter(organization__is_active=True)
        .only(
            '_is_deactivated',
            'id',
            'key',
        )
        .all()
    )
    serializer_class = serializers.Serializer
    permission_classes = [DevicePermission]
    schema = schema

    @classmethod
    def invalidate_get_device_cache(cls, instance, **kwargs):
        """Called from signal receiver which performs cache invalidation"""
        view = cls()
        view.get_object.invalidate(view, str(instance.pk))
        logger.debug(f'invalidated view cache for device ID {instance.pk}')

    @classmethod
    def invalidate_get_charts_cache(cls, instance, *args, **kwargs):
        if isinstance(instance, Device):
            pk = instance.id
        elif isinstance(instance, Metric):
            pk = instance.object_id
        elif isinstance(instance, Chart):
            pk = instance.metric.object_id
        cls._get_charts.invalidate(None, None, pk)

    def get(self, request, pk):
        # ensure valid UUID
        try:
            pk = str(uuid.UUID(pk))
        except ValueError:
            return Response({'detail': 'not found'}, status=404)
        self.instance = self.get_object(pk)
        response = super().get(request, pk)
        if not request.query_params.get('csv'):
            charts_data = dict(response.data)
            device_metrics_data = MonitoringDeviceDetailSerializer(self.instance).data
            return Response(
                {**device_metrics_data, **charts_data}, status=status.HTTP_200_OK
            )
        return response

    @cache_memoize(CACHE_TIMEOUT, args_rewrite=get_charts_args_rewrite)
    def _get_charts(self, request, *args, **kwargs):
        ct = ContentType.objects.get_for_model(Device)
        return Chart.objects.filter(
            metric__object_id=args[0], metric__content_type_id=ct.id
        ).select_related('metric')

    def _get_additional_data(self, request, *args, **kwargs):
        if request.query_params.get('status', False):
            return {'data': self.instance.data}
        return {}

    @cache_memoize(CACHE_TIMEOUT, args_rewrite=get_device_args_rewrite)
    def get_object(self, pk):
        return super().get_object()

    def post(self, request, pk):
        self.instance = self.get_object(pk)
        if self.instance._is_deactivated:
            # If the device is deactivated, do not accept data.
            # We don't use "Device.is_deactivated()" to avoid
            # generating query for the related config.
            raise Http404
        self.instance.data = request.data
        # validate incoming data
        try:
            self.instance.validate_data()
        except ValidationError as e:
            logger.info(e.message)
            return Response(e.message, status=status.HTTP_400_BAD_REQUEST)
        time_obj = request.query_params.get(
            'time', now().utcnow().strftime('%d-%m-%Y_%H:%M:%S.%f')
        )
        current = request.query_params.get('current', False)
        try:
            time = datetime.strptime(time_obj, '%d-%m-%Y_%H:%M:%S.%f').replace(
                tzinfo=UTC
            )
        except ValueError:
            return Response({'detail': _('Incorrect time format')}, status=400)
        # writing data is intensive, let's pass that to the background workers
        write_device_metrics.delay(
            str(self.instance.pk), self.instance.data, time=time_obj, current=current
        )
        device_metrics_received.send(
            sender=self.model,
            instance=self.instance,
            request=request,
            time=time,
            current=current,
        )
        return Response(None)


device_metric = DeviceMetricView.as_view()


class MonitoringGeoJsonLocationList(GeoJsonLocationList):
    serializer_class = MonitoringGeoJsonLocationSerializer
    queryset = (
        Location.objects.filter(devicelocation__isnull=False)
        .annotate(
            device_count=Count('devicelocation'),
            ok_count=Count(
                'devicelocation',
                filter=Q(devicelocation__content_object__monitoring__status='ok'),
            ),
            problem_count=Count(
                'devicelocation',
                filter=Q(devicelocation__content_object__monitoring__status='problem'),
            ),
            critical_count=Count(
                'devicelocation',
                filter=Q(devicelocation__content_object__monitoring__status='critical'),
            ),
            unknown_count=Count(
                'devicelocation',
                filter=Q(devicelocation__content_object__monitoring__status='unknown'),
            ),
        )
        .order_by('-created')
    )


monitoring_geojson_location_list = MonitoringGeoJsonLocationList.as_view()


class MonitoringLocationDeviceList(LocationDeviceList):
    serializer_class = MonitoringLocationDeviceSerializer

    def get_queryset(self):
        return super().get_queryset().select_related('monitoring').order_by('name')


monitoring_location_device_list = MonitoringLocationDeviceList.as_view()


class MonitoringNearbyDeviceList(
    DeviceKeyAuthenticationMixin, FilterByOrganizationManaged, ListAPIView
):
    serializer_class = MonitoringNearbyDeviceSerializer
    pagination_class = ListViewPagination
    filter_backends = [DjangoFilterBackend]
    filterset_class = MonitoringNearbyDeviceFilter
    permission_classes = []

    def get_queryset(self):
        qs = Device.objects.select_related('monitoring')
        location_lookup = Q(devicelocation__content_object_id=self.kwargs['pk'])
        device_key = self.request.query_params.get('key')
        if device_key:
            location_lookup &= Q(devicelocation__content_object__key=device_key)
        if not self.request.user.is_superuser and not device_key:
            qs = self.get_organization_queryset(qs)
        location = get_object_or_404(Location.objects, location_lookup)
        return (
            qs.exclude(id=self.kwargs['pk'])
            .filter(
                devicelocation__isnull=False,
            )
            .annotate(
                distance=Round(
                    Distance('devicelocation__location__geometry', location.geometry)
                )
            )
            .order_by('distance')
        )


monitoring_nearby_device_list = MonitoringNearbyDeviceList.as_view()


class MonitoringDeviceList(DeviceListCreateView):
    """Lists devices and their monitoring status (health status).

    Supports session authentication and token authentication.

    `NOTE:` The response does not include the information and health
    status of the specific metrics, this information can be retrieved in
    the detail endpoint of each device.
    """

    serializer_class = MonitoringDeviceListSerializer
    http_method_names = ['get', 'head', 'options']
    filter_backends = [DjangoFilterBackend]
    filterset_class = MonitoringDeviceFilter

    def get_queryset(self):
        return super().get_queryset().select_related('monitoring').order_by('name')


monitoring_device_list = MonitoringDeviceList.as_view()


class WifiSessionListView(ProtectedAPIMixin, FilterByOrganizationManaged, ListAPIView):
    queryset = WifiSession.objects.select_related(
        'device', 'wifi_client', 'device__organization', 'device__group'
    )
    organization_field = 'device__organization'
    filter_backends = [DjangoFilterBackend]
    pagination_class = ListViewPagination
    filterset_class = WifiSessionFilter
    serializer_class = WifiSessionSerializer


wifi_session_list = WifiSessionListView.as_view()


class WifiSessionDetailView(
    ProtectedAPIMixin, FilterByOrganizationManaged, RetrieveAPIView
):
    queryset = WifiSession.objects.select_related(
        'device', 'wifi_client', 'device__organization'
    )
    organization_field = 'device__organization'
    serializer_class = WifiSessionSerializer


wifi_session_detail = WifiSessionDetailView.as_view()



from datetime import timedelta
from unittest.mock import patch

from django.core.exceptions import ValidationError
from django.test import TransactionTestCase
from django.utils.timezone import now
from freezegun import freeze_time
from swapper import load_model

from ...device.tests import TestDeviceMonitoringMixin
from .. import settings as app_settings
from ..classes import ConfigApplied, Iperf3, Ping
from ..tasks import auto_create_config_check, auto_create_iperf3_check, auto_create_ping

Check = load_model('check', 'Check')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')
Device = load_model('config', 'device')
Notification = load_model('openwisp_notifications', 'Notification')


class TestModels(TestDeviceMonitoringMixin, TransactionTestCase):
    _PING = app_settings.CHECK_CLASSES[0][0]
    _CONFIG_APPLIED = app_settings.CHECK_CLASSES[1][0]
    _IPERF3 = app_settings.CHECK_CLASSES[2][0]

    def test_check_str(self):
        c = Check(name='Test check')
        self.assertEqual(str(c), c.name)

    def test_check_no_content_type(self):
        c = Check(name='Ping class check', check_type=self._PING)
        m = c.check_instance._get_metric()
        self.assertEqual(m, Metric.objects.first())

    def test_check_str_with_relation(self):
        obj = self._create_user()
        c = Check(name='Check', content_object=obj)
        expected = '{0} (User: {1})'.format(c.name, obj)
        self.assertEqual(str(c), expected)

    def test_check_class(self):
        with self.subTest('Test Ping check Class'):
            c = Check(name='Ping class check', check_type=self._PING)
            self.assertEqual(c.check_class, Ping)
        with self.subTest('Test Configuration Applied check Class'):
            c = Check(
                name='Configuration Applied class check',
                check_type=self._CONFIG_APPLIED,
            )
            self.assertEqual(c.check_class, ConfigApplied)
        with self.subTest('Test Iperf3 check Class'):
            c = Check(
                name='Iperf3 class check',
                check_type=self._IPERF3,
            )
            self.assertEqual(c.check_class, Iperf3)

    def test_base_check_class(self):
        path = 'openwisp_monitoring.check.classes.base.BaseCheck'
        c = Check(name='Base check', check_type=path)
        i = c.check_instance
        i.validate_params()
        with self.assertRaises(NotImplementedError):
            i.check()

    def test_check_instance(self):
        obj = self._create_device(organization=self._create_org())
        with self.subTest('Test Ping check instance'):
            c = Check(
                name='Ping class check',
                check_type=self._PING,
                content_object=obj,
                params={},
            )
            i = c.check_instance
            self.assertIsInstance(i, Ping)
            self.assertEqual(i.related_object, obj)
            self.assertEqual(i.params, c.params)
        with self.subTest('Test Configuration Applied check instance'):
            c = Check(
                name='Configuration Applied class check',
                check_type=self._CONFIG_APPLIED,
                content_object=obj,
                params={},
            )
            i = c.check_instance
            self.assertIsInstance(i, ConfigApplied)
            self.assertEqual(i.related_object, obj)
            self.assertEqual(i.params, c.params)

        with self.subTest('Test Iperf3 check instance'):
            c = Check(
                name='Iperf3 class check',
                check_type=self._IPERF3,
                content_object=obj,
                params={},
            )
            i = c.check_instance
            self.assertIsInstance(i, Iperf3)
            self.assertEqual(i.related_object, obj)
            self.assertEqual(i.params, c.params)

    def test_validation(self):
        with self.subTest('Test Ping check validation'):
            check = Check(name='Ping check', check_type=self._PING, params={})
            try:
                check.full_clean()
            except ValidationError as e:
                self.assertIn('device', str(e))
            else:
                self.fail('ValidationError not raised')
        with self.subTest('Test Configuration Applied check validation'):
            check = Check(
                name='Config check', check_type=self._CONFIG_APPLIED, params={}
            )
            try:
                check.full_clean()
            except ValidationError as e:
                self.assertIn('device', str(e))
            else:
                self.fail('ValidationError not raised')

    def test_auto_check_creation(self):
        self.assertEqual(Check.objects.count(), 0)
        d = self._create_device(organization=self._create_org())
        self.assertEqual(Check.objects.count(), 3)
        with self.subTest('Test AUTO_PING'):
            c1 = Check.objects.filter(check_type=self._PING).first()
            self.assertEqual(c1.content_object, d)
            self.assertEqual(self._PING, c1.check_type)
        with self.subTest('Test AUTO_CONFIG_CHECK'):
            c2 = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
            self.assertEqual(c2.content_object, d)
            self.assertEqual(self._CONFIG_APPLIED, c2.check_type)
        with self.subTest('Test AUTO_IPERF3'):
            c3 = Check.objects.filter(check_type=self._IPERF3).first()
            self.assertEqual(c3.content_object, d)
            self.assertEqual(self._IPERF3, c3.check_type)

    def test_device_deleted(self):
        self.assertEqual(Check.objects.count(), 0)
        d = self._create_device(organization=self._create_org())
        self.assertEqual(Check.objects.count(), 3)
        d.delete(check_deactivated=False)
        self.assertEqual(Check.objects.count(), 0)

    def test_config_modified_device_problem(self):
        self._create_admin()
        self.assertEqual(Check.objects.count(), 0)
        with freeze_time(
            now() - timedelta(minutes=app_settings.CONFIG_CHECK_INTERVAL + 10)
        ):
            self._create_config(status='modified', organization=self._create_org())
        d = Device.objects.first()
        d.monitoring.update_status('ok')
        self.assertEqual(Check.objects.count(), 3)
        self.assertEqual(Metric.objects.filter(object_id=d.id).count(), 0)
        self.assertEqual(AlertSettings.objects.count(), 0)
        check = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        with freeze_time(now() - timedelta(minutes=10)):
            check.perform_check()
        self.assertEqual(Metric.objects.filter(object_id=d.id).count(), 1)
        self.assertEqual(AlertSettings.objects.count(), 1)
        # Check needs to be run again without mocking time for threshold crossed
        check.perform_check()
        m = Metric.objects.filter(configuration='config_applied').first()
        self.assertEqual(m.content_object, d)
        self.assertEqual(m.key, 'config_applied')
        dm = d.monitoring
        dm.refresh_from_db()
        m.refresh_from_db(fields=['is_healthy', 'is_healthy_tolerant'])
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, False)
        self.assertEqual(dm.status, 'problem')
        self.assertEqual(Notification.objects.count(), 1)

    def test_config_error(self):
        """Test that ConfigApplied checks are skipped when device config status is errored."""
        self._create_admin()
        self.assertEqual(Check.objects.count(), 0)
        with freeze_time(
            now() - timedelta(minutes=app_settings.CONFIG_CHECK_INTERVAL + 10)
        ):
            self._create_config(status='error', organization=self._create_org())
        dm = Device.objects.first().monitoring
        dm.update_status('ok')
        self.assertEqual(Check.objects.count(), 3)
        self.assertEqual(Metric.objects.filter(object_id=dm.id).count(), 0)
        self.assertEqual(AlertSettings.objects.count(), 0)
        check = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        with freeze_time(now() - timedelta(minutes=10)):
            check.perform_check()
        # Check needs to be run again without mocking time for threshold crossed
        self.assertEqual(check.perform_check(), 0)
        self.assertEqual(Metric.objects.filter(object_id=dm.device_id).count(), 1)
        m = Metric.objects.filter(object_id=dm.device_id).first()
        self.assertEqual(AlertSettings.objects.count(), 1)
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'problem')
        self.assertEqual(Notification.objects.filter(actor_object_id=m.id).count(), 0)
        # Check config recovery
        dm.device.config.set_status_applied()
        # We are once again querying for the check to override the cached property check_instance
        check = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        # must be performed multiple times to trepass tolerance
        check.perform_check()
        with freeze_time(now() + timedelta(minutes=10)):
            check.perform_check()
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'ok')
        self.assertEqual(Notification.objects.filter(actor_object_id=m.id).count(), 1)

    @patch(
        'openwisp_monitoring.device.base.models.app_settings.CRITICAL_DEVICE_METRICS',
        [{'key': 'config_applied', 'field_name': 'config_applied'}],
    )
    def test_config_check_critical_metric(self):
        with freeze_time(
            now() - timedelta(minutes=app_settings.CONFIG_CHECK_INTERVAL + 6)
        ):
            self._create_config(status='modified', organization=self._create_org())
        self.assertEqual(Check.objects.count(), 3)
        d = Device.objects.first()
        dm = d.monitoring
        dm.update_status('ok')
        check = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        check.perform_check()
        self.assertEqual(Metric.objects.count(), 1)
        self.assertEqual(AlertSettings.objects.count(), 1)
        m = Metric.objects.first()
        self.assertTrue(dm.is_metric_critical(m))
        # must be executed twice to trepass the tolerance
        with freeze_time(now() - timedelta(minutes=6)):
            check.perform_check()
        check.perform_check()
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'critical')

    def test_no_duplicate_check_created(self):
        self._create_config(organization=self._create_org())
        self.assertEqual(Check.objects.count(), 3)
        d = Device.objects.first()
        auto_create_config_check.delay(
            model=Device.__name__.lower(),
            app_label=Device._meta.app_label,
            object_id=str(d.pk),
        )
        auto_create_ping.delay(
            model=Device.__name__.lower(),
            app_label=Device._meta.app_label,
            object_id=str(d.pk),
        )
        auto_create_iperf3_check.delay(
            model=Device.__name__.lower(),
            app_label=Device._meta.app_label,
            object_id=str(d.pk),
        )
        self.assertEqual(Check.objects.count(), 3)

    def test_device_unreachable_no_config_check(self):
        self._create_config(status='modified', organization=self._create_org())
        d = self.device_model.objects.first()
        d.monitoring.update_status('critical')
        self.assertEqual(Check.objects.count(), 3)
        c2 = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        c2.perform_check()
        self.assertEqual(Metric.objects.count(), 0)
        self.assertIsNone(c2.perform_check())

    def test_device_unknown_no_config_check(self):
        self._create_admin()
        self._create_config(status='modified', organization=self._create_org())
        d = self.device_model.objects.first()
        d.monitoring.update_status('unknown')
        self.assertEqual(Check.objects.count(), 3)
        c2 = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        c2.perform_check()
        self.assertEqual(Metric.objects.count(), 0)
        self.assertEqual(Notification.objects.count(), 0)
        self.assertIsNone(c2.perform_check())

    def test_device_organization_disabled_check_not_performed(self):
        self._create_config(
            status='modified', organization=self._create_org(is_active=False)
        )
        self.assertEqual(Check.objects.count(), 3)
        check = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        with patch(f'{self._CONFIG_APPLIED}.check') as mocked_check:
            check.perform_check()
        mocked_check.assert_not_called()

    def test_deactivated_device_check_not_performed(self):
        config = self._create_config(status='modified', organization=self._create_org())
        self.assertEqual(Check.objects.filter(is_active=True).count(), 3)
        config.device.deactivate()
        config.set_status_deactivated()
        check = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        with patch(f'{self._CONFIG_APPLIED}.check') as mocked_check:
            check.perform_check()
        mocked_check.assert_not_called()

    def test_config_check_problem_with_interval(self):
        self._create_admin()
        d = self._create_device(organization=self._create_org())
        self._create_config(device=d)
        check = Check.objects.filter(check_type=self._CONFIG_APPLIED).first()
        self.assertEqual(Check.objects.count(), 3)
        self.assertEqual(AlertSettings.objects.count(), 0)
        self.assertFalse(Metric.objects.filter(configuration='config_applied').exists())
        d.monitoring.update_status('ok')
        # Running check just after config modified
        check.perform_check()
        d.monitoring.refresh_from_db()
        self.assertTrue(Metric.objects.filter(configuration='config_applied').exists())
        metric = Metric.objects.get(configuration='config_applied', object_id=str(d.pk))
        # metric health should not change because the config
        # hasn't been changed for more than the interval
        self.assertTrue(metric.is_healthy)
        # Device monitoring status should be remain unchanged
        self.assertEqual(d.monitoring.status, 'ok')
        self.assertEqual(AlertSettings.objects.count(), 1)
        # Running check just when config was modified more than CONFIG_CHECK_INTERVAL ago
        with freeze_time(
            now() + timedelta(minutes=app_settings.CONFIG_CHECK_INTERVAL + 1)
        ):
            result = check.perform_check()
        d.monitoring.refresh_from_db()
        metric.refresh_from_db()
        self.assertEqual(result, 0)
        self.assertFalse(metric.is_healthy)
        self.assertEqual(d.monitoring.status, 'problem')



from django.core.exceptions import ImproperlyConfigured

from ..settings import get_settings_value


def get_critical_device_metrics():
    critical_metrics = get_settings_value(
        'CRITICAL_DEVICE_METRICS',
        [{'key': 'ping', 'field_name': 'reachable'}],
    )
    for item in critical_metrics:  # pragma: no cover
        try:
            assert 'key' in item
            assert 'field_name' in item
        except AssertionError as e:
            raise ImproperlyConfigured(
                'OPENWISP_MONITORING_CRITICAL_DEVICE_METRICS must contain the following keys: key, field_name'
            ) from e
    return critical_metrics


def get_health_status_labels():
    default_labels = {
        'unknown': 'unknown',
        'ok': 'ok',
        'problem': 'problem',
        'critical': 'critical',
        'deactivated': 'deactivated',
    }
    labels = default_labels.copy()
    configured_labels = get_settings_value(
        'HEALTH_STATUS_LABELS',
        default_labels,
    )
    labels.update(configured_labels)
    return labels


SHORT_RETENTION_POLICY = get_settings_value('SHORT_RETENTION_POLICY', '24h0m0s')
DEFAULT_RETENTION_POLICY = get_settings_value('DEFAULT_RETENTION_POLICY', '26280h0m0s')
CRITICAL_DEVICE_METRICS = get_critical_device_metrics()
HEALTH_STATUS_LABELS = get_health_status_labels()
AUTO_CLEAR_MANAGEMENT_IP = get_settings_value('AUTO_CLEAR_MANAGEMENT_IP', True)
# Triggers spontaneous recovery of device based on corresponding signals
DEVICE_RECOVERY_DETECTION = get_settings_value('DEVICE_RECOVERY_DETECTION', True)
MAC_VENDOR_DETECTION = get_settings_value('MAC_VENDOR_DETECTION', True)
DASHBOARD_MAP = get_settings_value('DASHBOARD_MAP', True)
WIFI_SESSIONS_ENABLED = get_settings_value('WIFI_SESSIONS_ENABLED', True)



from rest_framework import serializers
from swapper import load_model

from openwisp_controller.config.api.serializers import DeviceListSerializer
from openwisp_controller.geo.api.serializers import (
    GeoJsonLocationSerializer,
    LocationDeviceSerializer,
)
from openwisp_users.api.mixins import FilterSerializerByOrgManaged

Device = load_model('config', 'Device')
DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
DeviceData = load_model('device_monitoring', 'DeviceData')
Device = load_model('config', 'Device')
WifiSession = load_model('device_monitoring', 'WifiSession')
WifiClient = load_model('device_monitoring', 'WifiClient')


class BaseDeviceMonitoringSerializer(serializers.ModelSerializer):
    class Meta:
        model = DeviceMonitoring
        fields = ('status',)


class DeviceMonitoringLocationSerializer(BaseDeviceMonitoringSerializer):
    status_label = serializers.SerializerMethodField()

    def get_status_label(self, obj):
        return obj.get_status_display()

    class Meta(BaseDeviceMonitoringSerializer.Meta):
        fields = BaseDeviceMonitoringSerializer.Meta.fields + ('status_label',)


class DeviceMonitoringSerializer(BaseDeviceMonitoringSerializer):
    related_metrics = serializers.SerializerMethodField()

    def get_related_metrics(self, obj):
        return obj.related_metrics.values('name', 'is_healthy').order_by('name')

    class Meta(BaseDeviceMonitoringSerializer.Meta):
        fields = BaseDeviceMonitoringSerializer.Meta.fields + ('related_metrics',)


class MonitoringLocationDeviceSerializer(LocationDeviceSerializer):
    monitoring = DeviceMonitoringLocationSerializer()


class MonitoringNearbyDeviceSerializer(
    FilterSerializerByOrgManaged, serializers.ModelSerializer
):
    monitoring_status = serializers.CharField(source='monitoring.status')
    distance = serializers.SerializerMethodField('get_distance')
    monitoring_data = serializers.SerializerMethodField('get_monitoring_data')

    class Meta(DeviceListSerializer.Meta):
        model = Device
        fields = [
            'id',
            'name',
            'organization',
            'group',
            'mac_address',
            'management_ip',
            'model',
            'os',
            'system',
            'notes',
            'distance',
            'monitoring_status',
            'monitoring_data',
        ]

    def get_distance(self, obj):
        return obj.distance.m

    def get_monitoring_data(self, obj):
        return DeviceData.objects.only('id').get(id=obj.id).data


class MonitoringDeviceListSerializer(DeviceListSerializer):
    monitoring = BaseDeviceMonitoringSerializer(read_only=True)

    def get_status(self, obj):
        return obj.get_status_display()

    class Meta:
        model = Device
        fields = [
            'id',
            'name',
            'organization',
            'group',
            'mac_address',
            'key',
            'last_ip',
            'management_ip',
            'model',
            'os',
            'system',
            'notes',
            'config',
            'monitoring',
            'created',
            'modified',
        ]


class MonitoringDeviceDetailSerializer(MonitoringDeviceListSerializer):
    monitoring = DeviceMonitoringSerializer(read_only=True)


class MonitoringGeoJsonLocationSerializer(GeoJsonLocationSerializer):
    ok_count = serializers.IntegerField()
    problem_count = serializers.IntegerField()
    critical_count = serializers.IntegerField()
    unknown_count = serializers.IntegerField()


class WifiClientSerializer(serializers.ModelSerializer):
    wifi6 = serializers.CharField(source='he', read_only=True)
    wifi5 = serializers.CharField(source='vht', read_only=True)
    wifi4 = serializers.CharField(source='ht', read_only=True)

    class Meta:
        model = WifiClient
        fields = [
            'mac_address',
            'vendor',
            'wifi6',
            'wifi5',
            'wifi4',
            'wmm',
            'wds',
            'wps',
        ]


class WifiSessionSerializer(serializers.ModelSerializer):
    client = WifiClientSerializer(source='wifi_client')
    organization = serializers.CharField(source='device.organization', read_only=True)
    device = serializers.CharField(source='device.name', read_only=True)

    class Meta:
        model = WifiSession
        fields = [
            'id',
            'organization',
            'device',
            'ssid',
            'interface_name',
            'client',
            'start_time',
            'stop_time',
            'modified',
        ]



# Generated by Django 3.2.10 on 2022-01-05 17:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('monitoring', '0001_squashed_0023_alert_settings_tolerance_remove_default'),
    ]

    operations = [
        migrations.AddField(
            model_name='metric',
            name='is_healthy_tolerant',
            field=models.BooleanField(blank=True, default=None, null=True),
        ),
    ]



from celery import shared_task
from django.core.exceptions import ObjectDoesNotExist
from swapper import load_model

from openwisp_utils.tasks import OpenwispCeleryTask

from ..db import timeseries_db
from ..db.exceptions import TimeseriesWriteException
from .settings import RETRY_OPTIONS
from .signals import post_metric_write


def _metric_post_write(name, values, metric, check_threshold_kwargs=None, **kwargs):
    if not metric or not check_threshold_kwargs:
        return
    try:
        Metric = load_model('monitoring', 'Metric')
        if not isinstance(metric, Metric):
            metric = Metric.objects.select_related('alertsettings').get(pk=metric)
    except ObjectDoesNotExist:
        # The metric can be deleted by the time threshold is being checked.
        # This can happen as the task is being run async.
        pass
    else:
        metric.check_threshold(**check_threshold_kwargs)
        signal_kwargs = dict(
            sender=metric.__class__,
            metric=metric,
            values=values,
            time=kwargs.get('timestamp'),
            current=kwargs.get('current', 'False'),
        )
        post_metric_write.send(**signal_kwargs)


@shared_task(
    base=OpenwispCeleryTask,
    bind=True,
    autoretry_for=(TimeseriesWriteException,),
    **RETRY_OPTIONS
)
def timeseries_write(
    self, name, values, metric=None, check_threshold_kwargs=None, **kwargs
):
    """Writes and retries with exponential backoff on failures."""
    timeseries_db.write(name, values, **kwargs)
    _metric_post_write(name, values, metric, check_threshold_kwargs, **kwargs)


def _timeseries_write(name, values, metric=None, check_threshold_kwargs=None, **kwargs):
    """Handles writes synchronously when using UDP mode."""
    if timeseries_db.use_udp:
        func = timeseries_write
    else:
        func = timeseries_write.delay
        metric = metric.pk if metric else None
    func(
        name=name,
        values=values,
        metric=metric,
        check_threshold_kwargs=check_threshold_kwargs,
        **kwargs
    )


@shared_task(
    base=OpenwispCeleryTask,
    bind=True,
    autoretry_for=(TimeseriesWriteException,),
    **RETRY_OPTIONS
)
def timeseries_batch_write(self, data):
    """Writes data in batches.

    Similar to timeseries_write function above, but operates on list of
    metric data (batch operation)
    """
    timeseries_db.batch_write(data)
    for metric_data in data:
        _metric_post_write(**metric_data)


def _timeseries_batch_write(data):
    """If the timeseries database is using UDP to write data, then write data synchronously."""
    if timeseries_db.use_udp:
        timeseries_batch_write(data=data)
    else:
        for item in data:
            item['metric'] = item['metric'].pk
        timeseries_batch_write.delay(data=data)


@shared_task(base=OpenwispCeleryTask)
def delete_timeseries(key, tags):
    timeseries_db.delete_series(key=key, tags=tags)


@shared_task
def migrate_timeseries_database():
    """Performs migrations of timeseries datab.

    Performed asynchronously, due to changes introduced in
    https://github.com/openwisp/openwisp-monitoring/pull/368

    To be removed in a future release.
    """
    from .migrations.influxdb.influxdb_alter_structure_0006 import (
        migrate_influxdb_structure,
    )

    migrate_influxdb_structure()



from unittest.mock import patch

from django.core.exceptions import ValidationError
from django.test import TransactionTestCase
from swapper import load_model

from ... import settings as monitoring_settings
from ...device.tests import TestDeviceMonitoringMixin
from .. import settings
from ..classes import Ping
from ..classes.ping import get_ping_schema
from ..exceptions import OperationalError
from . import _FPING_REACHABLE, _FPING_UNREACHABLE

Chart = load_model('monitoring', 'Chart')
AlertSettings = load_model('monitoring', 'AlertSettings')
Metric = load_model('monitoring', 'Metric')
Check = load_model('check', 'Check')


class TestPing(TestDeviceMonitoringMixin, TransactionTestCase):
    _PING = settings.CHECK_CLASSES[0][0]
    _RESULT_KEYS = ['reachable', 'loss', 'rtt_min', 'rtt_avg', 'rtt_max']
    _RTT_KEYS = _RESULT_KEYS[-3:]
    _UNRECOGNIZED_OUTPUT = (
        '',
        bytes("fping: option requires an argument -- 'z'", encoding='utf8'),
    )

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    def test_check_ping_no_params(self, mocked_method):
        device = self._create_device(organization=self._create_org())
        # will ping localhost
        device.management_ip = '127.0.0.1'
        check = Check(
            name='Ping check', check_type=self._PING, content_object=device, params={}
        )
        result = check.perform_check(store=False)
        for key in self._RESULT_KEYS:
            self.assertIn(key, result)
        self.assertEqual(result['reachable'], 1)
        self.assertEqual(result['loss'], 0.0)
        for key in self._RTT_KEYS:
            self.assertTrue(result[key] < 1)

    def test_check_ping_params(self):
        device = self._create_device(organization=self._create_org())
        # will ping localhost
        device.management_ip = '127.0.0.1'
        check = Check(
            name='Ping check',
            check_type=self._PING,
            content_object=device,
            params={'count': 2, 'interval': 10, 'bytes': 12, 'timeout': 50},
        )
        result = check.perform_check(store=False)
        for key in self._RESULT_KEYS:
            self.assertIn(key, result)
        self.assertEqual(result['reachable'], 1)
        self.assertEqual(result['loss'], 0.0)
        for key in self._RTT_KEYS:
            self.assertTrue(result[key] < 1)

    @patch.object(
        settings,
        'PING_CHECK_CONFIG',
        {
            'timeout': {'default': '10000'},
            'count': {'default': 22},
            'bytes': {'default': 1024},
        },
    )
    def test_ping_check_config(self, *args):
        with patch.object(Ping, 'schema', get_ping_schema()):
            device = self._create_device(organization=self._create_org())
            # will ping localhost
            device.management_ip = '127.0.0.1'
            check = Check(
                name='Ping check',
                check_type=self._PING,
                content_object=device,
                params={},
            )
            with patch.object(
                Ping, '_command', return_value=_FPING_REACHABLE
            ) as mocked_command:
                check.perform_check(store=False)
            mocked_command.assert_called_once_with(
                [
                    'fping',
                    '-e',
                    '-c 22',
                    '-p 25',
                    '-b 1024',
                    '-t 10000',
                    '-q',
                    '127.0.0.1',
                ]
            )

    @patch.object(Ping, '_command', return_value=_FPING_UNREACHABLE)
    def test_check_ping_unreachable(self, mocked_method):
        device = self._create_device(organization=self._create_org())
        device.management_ip = '192.168.255.255'
        check = Check(
            name='Ping check',
            check_type=self._PING,
            content_object=device,
            params={'timeout': 50, 'count': 3},
        )
        result = check.perform_check(store=False)
        for key in self._RESULT_KEYS[0:2]:
            self.assertIn(key, result)
        self.assertFalse(result['reachable'], 0)
        self.assertEqual(result['loss'], 100.0)

    @patch.object(Ping, '_command', return_value=_UNRECOGNIZED_OUTPUT)
    def test_operational_error(self, _command):
        device = self._create_device(organization=self._create_org())
        device.management_ip = '127.0.0.1'
        check = Check(
            name='Ping check',
            check_type=self._PING,
            content_object=device,
            params={'timeout': 50, 'count': 3},
        )
        try:
            check.perform_check(store=False)
        except OperationalError as e:
            self.assertIn('Unrecognized fping output:', str(e))
        else:
            self.fail('OperationalError not raised')

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    def _check_no_ip_case(self, status, mocked_method, management_ip_only=False):
        device = self._create_device(
            organization=self._create_org(), last_ip='127.0.0.1'
        )
        device.monitoring.update_status(status)
        check = Check(
            name='Ping check', check_type=self._PING, content_object=device, params={}
        )
        result = check.perform_check(store=True)
        if not management_ip_only:
            if status != 'unknown':
                expected_result = {'reachable': 0, 'loss': 100}
                expected_status = 'critical'
                expected_metrics_count = 1
            else:
                expected_result = None
                expected_status = status
                expected_metrics_count = 0
            self.assertEqual(result, expected_result)
        else:
            expected_status = 'ok'
            expected_metrics_count = 1
        device.monitoring.refresh_from_db()
        self.assertEqual(device.monitoring.status, expected_status)
        self.assertEqual(Metric.objects.count(), expected_metrics_count)

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    @patch('openwisp_monitoring.check.settings.MANAGEMENT_IP_ONLY', True)
    def test_device_without_ip_unknown_status(self, mocked_method):
        self._check_no_ip_case('unknown')

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    @patch('openwisp_monitoring.check.settings.MANAGEMENT_IP_ONLY', True)
    def test_device_without_ip_ok_status(self, mocked_method):
        self._check_no_ip_case('ok')

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    @patch('openwisp_monitoring.check.settings.MANAGEMENT_IP_ONLY', True)
    def test_device_without_ip_problem_status(self, mocked_method):
        self._check_no_ip_case('problem')

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    @patch('openwisp_monitoring.check.settings.MANAGEMENT_IP_ONLY', True)
    def test_device_without_ip_critical_status(self, mocked_method):
        self._check_no_ip_case('critical')

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    @patch('openwisp_monitoring.check.settings.MANAGEMENT_IP_ONLY', False)
    def test_device_with_last_ip_unknown_status(self, mocked_method):
        self._check_no_ip_case('unknown', management_ip_only=True)

    def test_content_object_none(self):
        check = Check(name='Ping check', check_type=self._PING, params={})
        try:
            check.check_instance.validate()
        except ValidationError as e:
            self.assertIn('device', str(e))
        else:
            self.fail('ValidationError not raised')

    def test_content_object_not_device(self):
        check = Check(
            name='Ping check',
            check_type=self._PING,
            content_object=self._create_user(),
            params={},
        )
        try:
            check.check_instance.validate()
        except ValidationError as e:
            self.assertIn('device', str(e))
        else:
            self.fail('ValidationError not raised')

    def test_schema_violation(self):
        device = self._create_device(organization=self._create_org())
        device.management_ip = '127.0.0.1'
        invalid_params = [
            {'count': 1},
            {'interval': 9},
            {'bytes': 0},
            {'timeout': 2},
            {'count': 'ciao'},
            {'wrong': True},
            {'bytes': 999999, 'count': 99},
        ]
        for params in invalid_params:
            check = Check(
                name='Ping check',
                check_type=self._PING,
                content_object=device,
                params=params,
            )
            try:
                check.check_instance.validate()
            except ValidationError as e:
                self.assertIn('Invalid param', str(e))
            else:
                self.fail('ValidationError not raised')

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    def test_store_result(self, mocked_method):
        self.assertEqual(Check.objects.count(), 0)
        device = self._create_device(organization=self._create_org())
        device.management_ip = '10.40.0.1'
        device.save()
        # check created automatically by autoping
        self.assertEqual(Check.objects.count(), 3)
        self.assertEqual(Metric.objects.count(), 0)
        self.assertEqual(Chart.objects.count(), 0)
        self.assertEqual(AlertSettings.objects.count(), 0)
        check = Check.objects.filter(check_type=self._PING).first()
        result = check.perform_check()
        self.assertEqual(Metric.objects.count(), 1)
        self.assertEqual(Chart.objects.count(), 3)
        self.assertEqual(AlertSettings.objects.count(), 1)
        m = Metric.objects.first()
        self.assertEqual(m.content_object, device)
        self.assertEqual(m.key, 'ping')
        points = self._read_metric(m, limit=None, extra_fields=list(result.keys()))
        self.assertEqual(len(points), 1)
        self.assertEqual(points[0]['reachable'], result['reachable'])
        self.assertEqual(points[0]['loss'], result['loss'])
        self.assertEqual(points[0]['rtt_min'], result['rtt_min'])
        self.assertEqual(points[0]['rtt_avg'], result['rtt_avg'])
        self.assertEqual(points[0]['rtt_max'], result['rtt_max'])

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    @patch.object(monitoring_settings, 'AUTO_CHARTS', return_value=[])
    def test_auto_chart_disabled(self, *args):
        device = self._create_device(organization=self._create_org())
        device.last_ip = '127.0.0.1'
        device.save()
        check = Check.objects.first()
        self.assertEqual(Chart.objects.count(), 0)
        check.perform_check()
        self.assertEqual(Chart.objects.count(), 0)



from django.utils.translation import gettext_lazy as _
from swapper import load_model

from openwisp_users.multitenancy import MultitenantOrgFilter
from openwisp_utils.admin_theme.filters import AutocompleteFilter

Device = load_model('config', 'Device')


class DeviceOrganizationFilter(MultitenantOrgFilter):
    rel_model = Device
    parameter_name = 'device__organization'


class DeviceGroupFilter(AutocompleteFilter):
    field_name = 'group'
    parameter_name = 'device__group'
    title = _('group')
    rel_model = Device


class DeviceFilter(AutocompleteFilter):
    field_name = 'device'
    parameter_name = 'device'
    title = _('device')



from django.urls import path, re_path

from . import views

app_name = 'monitoring'

urlpatterns = [
    path(
        'api/v1/monitoring/device/',
        views.monitoring_device_list,
        name='api_monitoring_device_list',
    ),
    re_path(
        r'^api/v1/monitoring/device/(?P<pk>[^/]+)/$',
        views.device_metric,
        name='api_device_metric',
    ),
    re_path(
        r'^api/v1/monitoring/device/(?P<pk>[^/]+)/nearby-devices/$',
        views.monitoring_nearby_device_list,
        name='api_monitoring_nearby_device_list',
    ),
    path(
        'api/v1/monitoring/geojson/',
        views.monitoring_geojson_location_list,
        name='api_location_geojson',
    ),
    re_path(
        r'^api/v1/monitoring/location/(?P<pk>[^/]+)/device/$',
        views.monitoring_location_device_list,
        name='api_location_device_list',
    ),
    path(
        'api/v1/monitoring/wifi-session/',
        views.wifi_session_list,
        name='api_wifi_session_list',
    ),
    re_path(
        r'^api/v1/monitoring/wifi-session/(?P<pk>[^/]+)/$',
        views.wifi_session_detail,
        name='api_wifi_session_detail',
    ),
]



from unittest.mock import patch

from django.apps import apps
from django.test import TestCase
from influxdb import InfluxDBClient
from requests.exceptions import ConnectionError

from openwisp_monitoring.settings import MONITORING_TIMESERIES_RETRY_OPTIONS


class TestDatabase(TestCase):
    app = 'monitoring'

    @patch.object(InfluxDBClient, 'create_database', side_effect=ConnectionError())
    @patch('openwisp_monitoring.utils.sleep')
    def test_check_retry(self, sleep_mock, mock):
        max_retries = MONITORING_TIMESERIES_RETRY_OPTIONS.get('max_retries')
        delay = MONITORING_TIMESERIES_RETRY_OPTIONS.get('delay')
        with patch('logging.Logger.info') as mocked_logger:
            with self.assertRaises(ConnectionError):
                apps.get_app_config(self.app).ready()
            self.assertEqual(mocked_logger.call_count, max_retries)
            mocked_logger.assert_called_with(
                'Error while executing method "create_database":'
                f'\n\nAttempt {max_retries} out of {max_retries}.\n'
            )
        self.assertEqual(mock.call_count, max_retries)
        self.assertEqual(sleep_mock.call_count, max_retries - 3)
        sleep_mock.assert_called_with(delay)



from django.test import TestCase

from openwisp_utils.admin_theme.dashboard import DASHBOARD_CHARTS, DASHBOARD_TEMPLATES


class TestApps(TestCase):
    def test_monitoring_status_chart_registered(self):
        chart_config = DASHBOARD_CHARTS.get(0, None)
        self.assertIsNotNone(chart_config)
        self.assertEqual(chart_config['name'], 'Monitoring Status')
        self.assertIn('labels', chart_config)
        self.assertNotIn('filters', chart_config)
        query_params = chart_config['query_params']
        self.assertIn('group_by', query_params)
        self.assertEqual(query_params['group_by'], 'monitoring__status')

    def test_device_map_template_registered(self):
        template_config = DASHBOARD_TEMPLATES.get(0, None)
        self.assertIsNotNone(template_config)
        self.assertEqual(
            template_config[0]['template'], 'admin/dashboard/device_map.html'
        )
        self.assertEqual(
            template_config[0].get('css'),
            (
                'monitoring/css/device-map.css',
                'leaflet/leaflet.css',
                'monitoring/css/leaflet.fullscreen.css',
                'monitoring/css/netjsongraph.css',
            ),
        )
        self.assertEqual(
            template_config[0].get('js'),
            (
                'monitoring/js/lib/netjsongraph.min.js',
                'monitoring/js/lib/leaflet.fullscreen.min.js',
                'monitoring/js/device-map.js',
            ),
        )



# Manually created
from django.db import migrations


def forward_migration(apps, schema_editor):
    from ..tasks import migrate_timeseries_database

    migrate_timeseries_database.delay()


class Migration(migrations.Migration):
    dependencies = [('monitoring', '0005_migrate_metrics')]

    operations = [
        migrations.RunPython(forward_migration, reverse_code=migrations.RunPython.noop)
    ]



from django.core.management.base import BaseCommand

from ...tasks import migrate_timeseries_database


class BaseMigrateTimeseriesCommand(BaseCommand):
    help = 'Migrate timeseries database asynchronously'

    def handle(self, *args, **options):
        migrate_timeseries_database.delay()
